﻿#include "editor.h"
#include "bridge.h"
#include "../common/sources/system/cgraphics2d.h"
#include "tools.h"
#include <QDebug>

CEditor  editor;

//---------------------------------------------------------------------------------------------
CEditor::CEditor()
{
    mUndoStack = new QUndoStack();
    mSelectStart.x = -1;
}

//---------------------------------------------------------------------------------------------
// trouve le coté du mur pour une position relative à une case
EWallSide   CEditor::getWallSide(CVec2 localPos)
{
    bool side1 = (localPos.y < localPos.x);
    bool side2 = (localPos.y < TILE_SIZE - 1 - localPos.x);

    if (side1 == true && side2 == true)
        return eWallSideUp;

    if (side1 == true && side2 == false)
        return eWallSideRight;

    if (side1 == false && side2 == true)
        return eWallSideLeft;

    return eWallSideDown;
}

//---------------------------------------------------------------------------------------------
// trouve le coté du mur pour la position de la souris dans la map
EWallSide   CEditor::getWallSideAbs(CVec2 pos)
{
    return getWallSide(pos % TILE_SIZE);
}

//---------------------------------------------------------------------------------------------
// dessine un rectangle entourant une case pour le curseur et la sélection
void   CEditor::drawTileRect(QImage* image, CVec2 mousePos, QColor color)
{
    CVec2   pos = mousePos / TILE_SIZE;
    CVec2   pos2 = pos * TILE_SIZE;

    CRect   r(CVec2(pos2), CVec2(pos2.x + TILE_SIZE - 1, pos2.y + TILE_SIZE - 1));
    graph2D.rect(image, r, color, false);
}

//---------------------------------------------------------------------------------------------
// dessine un rectangle entourant un mur pour le curseur et la sélection
void   CEditor::drawWallRect(QImage* image, CVec2 mousePos, QColor color)
{
    CVec2   pos = mousePos / TILE_SIZE;
    CVec2   pos2 = pos * TILE_SIZE;
    CRect   r;

    CVec2   localPos = mousePos - pos2;
    EWallSide   side = getWallSide(localPos);

    switch (side)
    {
        case eWallSideUp:
            r = CRect(pos2, CVec2(pos2.x + TILE_SIZE - 1, pos2.y + TILE_SIZE / 4 - 1));
            break;

        case eWallSideDown:
            r = CRect(CVec2(pos2.x, pos2.y + (TILE_SIZE * 3) / 4),
                      CVec2(pos2.x + TILE_SIZE - 1, pos2.y + TILE_SIZE - 1));
            break;

        case eWallSideLeft:
            r = CRect(pos2, CVec2(pos2.x + TILE_SIZE / 4 - 1, pos2.y + TILE_SIZE - 1));
            break;

        case eWallSideRight:
            r = CRect(CVec2(pos2.x + (TILE_SIZE * 3) / 4, pos2.y),
                      CVec2(pos2.x + TILE_SIZE - 1, pos2.y + TILE_SIZE - 1));
            break;

        default:
            break;
    }

    graph2D.rect(image, r, color, false);
}

//---------------------------------------------------------------------------------------------
// dessine un rectangle entourant un objet pour le curseur et la sélection
void   CEditor::drawObjectRect(QImage* image, CVec2 mousePos, QColor color)
{
    CRect   r;

    CVec2   localPos = mousePos;
    localPos /= TILE_SIZE / 2;
    localPos *= TILE_SIZE / 2;

    r = CRect(localPos, CVec2(localPos.x + TILE_SIZE / 2 - 1, localPos.y + TILE_SIZE / 2 - 1));
    graph2D.rect(image, r, color, false);
}

//---------------------------------------------------------------------------------------------
void CEditor::drawCursor(QImage* image, CVec2 mousePos, ETabIndex tabIndex)
{
    if (mousePos.x != -1)
    {
        if (tabIndex == eTabTiles)
            drawTileRect(image, mousePos, CURSOR_COLOR);
        else if (tabIndex == eTabWalls)
            drawWallRect(image, mousePos, CURSOR_COLOR);
        else if (tabIndex == eTabObjects)
            drawObjectRect(image, mousePos, CURSOR_COLOR);
    }
}

//---------------------------------------------------------------------------------------------
// dessine la sélection quand on est en mode case
void   CEditor::drawTileSelection(QImage* image)
{
    CVec2   startPos = mSelectStart / TILE_SIZE;
    CVec2   endPos   = mSelectEnd   / TILE_SIZE;

    if (startPos == endPos)
    {
        drawTileRect(image, mSelectStart, SELECTION_COLOR);
    }
    else
    {
        CRect   sel = getSelectionRect();
        drawRectangleSelection(image, sel);
    }
}

//---------------------------------------------------------------------------------------------
// dessine la sélection quand on est en mode mur
void   CEditor::drawWallSelection(QImage* image)
{
    CVec2   startPos = mSelectStart / TILE_SIZE;
    CVec2   endPos   = mSelectEnd   / TILE_SIZE;
    EWallSide   startSide = getWallSideAbs(mSelectStart);
    EWallSide   endSide   = getWallSideAbs(mSelectEnd);

    if (startPos == endPos && startSide == endSide)
    {
        drawWallRect(image, mSelectStart, SELECTION_COLOR);
    }
    else
    {
        CRect   sel = getSelectionRect();
        drawRectangleSelection(image, sel);
    }
}

//---------------------------------------------------------------------------------------------
// dessine la sélection quand on est en mode objet
void   CEditor::drawObectSelection(QImage* image)
{
    CVec2   startPos = mSelectStart / (TILE_SIZE / 2);
    CVec2   endPos   = mSelectEnd   / (TILE_SIZE / 2);

    if (startPos == endPos)
    {
        drawObjectRect(image, mSelectStart, SELECTION_COLOR);
    }
    else
    {
        CRect   sel = getSelectionRect();
        drawRectangleSelection(image, sel);
    }
}

//---------------------------------------------------------------------------------------------
CRect CEditor::getSelectionRect()
{
    CVec2   startPos = mSelectStart / TILE_SIZE;
    CVec2   endPos = mSelectEnd / TILE_SIZE;
    CRect   rect;

    // gère les cas où on ne trace pas le rectangle de haut/gauche vers bas/droite
    rect.tl.x = MIN(startPos.x, endPos.x);
    rect.tl.y = MIN(startPos.y, endPos.y);
    rect.br.x = MAX(startPos.x, endPos.x);
    rect.br.y = MAX(startPos.y, endPos.y);
    return rect;
}

//---------------------------------------------------------------------------------------------
// dessine la séléction quand on sélectionne une zone rectangulaire
void   CEditor::drawRectangleSelection(QImage* image, CRect rect)
{
    rect.tl *= TILE_SIZE;
    rect.br *= TILE_SIZE;
    rect.br += CVec2(TILE_SIZE - 1, TILE_SIZE - 1);
    graph2D.rect(image, rect, SELECTION_COLOR, false);
}

//---------------------------------------------------------------------------------------------
void CEditor::drawSelection(QImage* image, ETabIndex tabIndex)
{
    if (mSelectStart.x != -1)
    {
        if (tabIndex == eTabTiles)
            drawTileSelection(image);
        else if (tabIndex == eTabWalls)
            drawWallSelection(image);
        else if (tabIndex == eTabObjects)
            drawObectSelection(image);
    }
}

//---------------------------------------------------------------------------------------------
void CEditor::drawPasteSelection(QImage* image)
{
    if (bridge.mTool == eQMLToolPaste &&
        bridge.mMousePos.x != -1 &&
        mClipboard.isEmpty() == false)
    {
        CVec2   pos1 = bridge.mMousePos / TILE_SIZE;
        CVec2   pos2 = pos1 + mClipboard.mSize - CVec2(1, 1);

        if (pos2.x >= map.mSize.x)
            pos2.x = map.mSize.x - 1;

        if (pos2.y >= map.mSize.y)
            pos2.y = map.mSize.y - 1;

        drawRectangleSelection(image, CRect(pos1, pos2));
    }
}


//---------------------------------------------------------------------------------------------
QImage CEditor::displayMap(CMap& map, CVec2 mousePos, ETabIndex tabIndex)
{
    QImage image(map.mSize.x * TILE_SIZE,
                 map.mSize.y * TILE_SIZE,
                 QImage::Format_RGBA8888);

    image.fill(QColor(0, 0, 0));

    QImage  tileGfx("tiles/Tile.png");
    QImage  wallGfx("tiles/Wall.png");

    if (map.isLoaded() == true)
    {
        // dessine les cases
        for (int y = 0; y < map.mSize.y; ++y)
            for (int x = 0; x < map.mSize.x; ++x)
            {
                CVec2   pos(x, y);
                CVec2   pos2 = pos * TILE_SIZE;

                if (map.getTile(pos)->mType != 0)
                    graph2D.drawImage(&image, pos2, tileGfx);
            }

        // dessine les murs
        for (int y = 0; y < map.mSize.y; ++y)
            for (int x = 0; x < map.mSize.x; ++x)
            {
                CVec2   pos(x, y);
                CVec2   pos2 = pos * TILE_SIZE;
                CTile*  tile = map.getTile(pos);

                if (tile->mWalls[eWallSideUp].mType != 0)
                    graph2D.drawImage(&image, pos2, wallGfx);

                if (tile->mWalls[eWallSideDown].mType != 0)
                    graph2D.drawImage(&image, pos2, wallGfx, 180);

                if (tile->mWalls[eWallSideLeft].mType != 0)
                    graph2D.drawImage(&image, pos2, wallGfx, -90);

                if (tile->mWalls[eWallSideRight].mType != 0)
                    graph2D.drawImage(&image, pos2, wallGfx, 90);
            }
    }

    drawSelection(&image, tabIndex);
    drawPasteSelection(&image);
    drawCursor(&image, mousePos, tabIndex);

    return image;
}

//---------------------------------------------------------------------------------------------
void    CEditor::useTool()
{
    switch (bridge.mTool)
    {
    case eQMLToolDraw:
        toolDraw();
        break;

    case eQMLToolSelect:
        toolSelect();
        break;

    case eQMLToolPaste:
        toolPaste();
        break;
    }
}

//---------------------------------------------------------------------------------------------
void    CEditor::toolDraw()
{
    if (bridge.rightPressed == true || bridge.leftPressed == true)
    {
        if (bridge.mTabIndex == eTabTiles)
        {
            // cases
            drawTileTool*   tool = new drawTileTool();
            int type = 0;

            if (bridge.rightPressed == true)
                type = 0;
            else if (bridge.leftPressed == true)
                type = 1;

            if (tool->init(bridge.mMousePos, type) == true)
            {
                unselect();
                mUndoStack->push(tool);
            }
            else
                delete tool;
        }
        else if (bridge.mTabIndex == eTabWalls)
        {
            // murs
            drawWallTool*   tool = new drawWallTool();
            int type = 0;

            if (bridge.rightPressed == true)
                type = 0;
            else if (bridge.leftPressed == true)
                type = 1;

            if (tool->init(bridge.mMousePos, type) == true)
            {
                unselect();
                mUndoStack->push(tool);
            }
            else
                delete tool;
        }
    }
}

//---------------------------------------------------------------------------------------------
void        CEditor::copyData(CRect rect)
{
    CVec2   size = rect.br - rect.tl + CVec2(1, 1);
    mClipboard.copyFromMap(rect.tl, size);
}

//---------------------------------------------------------------------------------------------
void        CEditor::toolCopy()
{
    if (mSelectStart.x != -1)
    {
        CRect   rect = getSelectionRect();
        copyData(rect);
        unselect();
    }
}

//---------------------------------------------------------------------------------------------
void        CEditor::toolCut()
{
    if (mSelectStart.x != -1)
    {
        cutTool*   tool = new cutTool();
        CRect   rect = getSelectionRect();
        copyData(rect);

        if (tool->init(mSelectStart / TILE_SIZE, bridge.mTabIndex) == true)
            mUndoStack->push(tool);
        else
            delete tool;
    }
    unselect();
}

//---------------------------------------------------------------------------------------------
void        CEditor::toolPaste()
{
    if (bridge.leftPressing == true)
    {
        unselect();
        pasteTool*   tool = new pasteTool();

        if (tool->init(bridge.mMousePos / TILE_SIZE, bridge.mTabIndex) == true)
            mUndoStack->push(tool);
        else
            delete tool;
    }
}

//---------------------------------------------------------------------------------------------
void    CEditor::unselect()
{
    mSelectStart.x = -1;
    bridge.updateQMLImage();
}

//---------------------------------------------------------------------------------------------
void    CEditor::toolSelect()
{
    if (bridge.leftPressed == true)
    {
        mSelectEnd = bridge.mMousePos;

        if (bridge.leftPressing == true)
            mSelectStart = mSelectEnd;

        bridge.updateQMLImage();
    }
}
